
package PointSet;
use strict 'vars';
use warnings;
use List::Util qw(max);
use Text::Balanced qw(extract_bracketed);
use FilterUtil qw(run_command bracket_list read_file);
use FilterConfig;

use FacetList;

# constructors

sub new {
  my $invokant=shift;
  my $class=ref $invokant || $invokant;
  my %arg=@_;

  die "must define rank, elements, points" 
    unless (defined($arg{rank}) && defined($arg{elements}) 
	    && defined($arg{points}));
  

  return bless ( { ELEMENTS=>$arg{elements},
		   RANK=>$arg{rank},
		   POINTS=> $arg{points} },$class);
}

sub input($$$$){
  my %arg=@_;
  my $file=$arg{file} || die "file must be specified";
  local($_)=$arg{type};
  
  return read_topcom($file) if ( m/^top/); 
  return read_lrs($file) if ( m/^(lrs|ine|cdd)/); 
  return from_poly(read_file($file)) if (m/^poly/);
  return read_raw($file);
}

sub read_raw($){
  return from_raw(read_file(shift));
}


sub read_lrs($){
  return from_lrs(read_file(shift));
}

sub read_topcom($){
  return from_topcom(read_file(shift));
}


# methods
sub elements{
  my $self=shift;
  return $self->{ELEMENTS};
}

sub rank{
  my $self=shift;
  return $self->{RANK};
}

sub extremes($){
  my $self=shift;

  my @lines=run_command($REDUND_CMD,$self->ine());

  return from_lrs(@lines);
}

#format as string for lrs/cdd	     
sub ine($@){
  my $self=shift;
  my @options=@_;

  my @lines;
  push(@lines, "V-representation", "begin", 
       sprintf("%d %d integer",
	       $self->elements(),
	       $self->rank()));
  push(@lines,map( join(" ", 1,@$_) , @{$self->{POINTS}}));
  push(@lines,"end");
  
  return join("\n",@lines,@options)."\n";
}

# format as points for topcom

sub topcom($){
  my $self=shift;
  return bracket_list('[',']',",\n",
		      map( bracket_list('[',']',
					",",@$_), @{$self->{POINTS}}));

}

sub facets($){
  my $self=shift;

  my @lines=run_command($LRS_CMD,$self->ine("incidence"));

  
  return extract_lrs_facets(@lines);
    
}

sub chirotope($){
  my $self=shift;
  my @lines= run_command("points2chiro",$self->topcom());
  
  return Chirotope::from_pm(@lines);
}

######################################################################
# utility routines. mostly work with arrays of lines repsenting files.

sub extract_lrs_facets(@){
  my @lines=@_;
  my @facets;
  my $dim;

  my $elements=0;

  foreach my $line (@lines){
    if ($line =~ m/^F\#(\d+)/){
      my $fnum=$1;
      $line =~ s|^.*rays||;
      print STDERR $line;
      $line =~ s/I.*$//;
      print STDERR $line;
      $line =~ s/\b\d+\*//;
      print STDERR $line;
      my ($gens,$extras)=split(":",$line);

      my @gens=split(" ",$gens);
      my @extras=();

      @extras=split(" ",$extras) if (defined($extras));
      $dim=scalar(@gens);
      my @nums=sort { $a <=> $b } (@gens,@extras);
      $elements=max($elements,$nums[$#nums]);
      print STDERR join(",",@nums)."\t".$line."\n";
      push(@facets,[ @nums ]);
    }
  }
  return new FacetList (rank=>$dim+1,
			elements=>$elements,
			facets=>[@facets]);
}



sub from_lrs(@){
  my @lines=@_;
  my @points=();
  
  shift(@lines) while (scalar(@lines) && $lines[0] !~ /^\s*begin\s*$/);
  die "garbled lrs output : @lines" if (!scalar(@lines));

  shift(@lines); # skip begin line
  my $params=shift(@lines);

  my ($m,$r)= ($params=~ m/\s*(\d+)\s+(\d+)/);
  
  foreach my $line (@lines){
    last if ($line =~ m/^\s*end\s*/);
    my @nums=split(" ",$line);
    die ("Ack. not an affine point? $line") if ($nums[0] != 1);
    shift(@nums);
    push(@points, [ @nums ]);
  }
  return new PointSet (elements=>$m, 
		       rank=>($r),
		       points=>[@points]);

}


sub from_topcom(@){
  my $str=join('',@_);
  local($_);
  my @points;
  my $elements=0;
  my $rank=0;

  $_=extract_bracketed($str,'[]');

  s/\s//g;
  s/\[\[//;
  s/\]\]//;

  my @lines=split(/\],\[/);

  foreach (@lines){
    chomp();
    s/[\[\]]//g;

    my @nums=split(",");
    $rank=scalar(@nums)+1;
    $elements++;
    push(@points,[@nums])
  }
  return new PointSet (rank=>$rank, elements=>$elements,
		       points=>[@points]);
}

sub from_poly(@){
  my @lines=@_;
  my @points;
  shift(@lines) while ($lines[0] !~ /^VERTICES$/);
  shift(@lines);

  foreach my $line (@lines){
    last if ($line =~ m/^\s*$/);
    my @nums=split(" ",$line);
    die ("Ack. not an affine point? $line") if ($nums[0] != 1);
    shift(@nums);
    push(@points, [ @nums ]);
  }
  return new PointSet (elements=>scalar(@points), 
		       rank=>(scalar(@{$points[0]})+1),
		       points=>[@points]);

}

sub from_raw(@){
  my @lines=@_;

  my @points;
  my $params=shift(@lines);
  
  my ($elements,$dim)=split(" ",$params);

  
  foreach my $line (@lines){
    push(@points,[split(" ",$line)]);
  }

  return new PointSet (elements=>$elements, 
		       rank=>($dim+1),
		       points=>[@points]);
}



1;
