/* z_lb.c (ZRAM: General load balancing)
   Ambros Marzetta
   95-09-20

 * $Source$
 * $Revision: 3145 $
 * $Date: 2004-08-25 18:01:45 -0300 (Wed, 25 Aug 2004) $
 */

#define _Z_LB_C_
#include <stdio.h>
#include <stdlib.h>

#include "z_lb.h"
#include "z_par.h"
#include "z_mp.h"
#include "s_seq.h"
#include "z_io.h"
#include "z_time.h"
#include "z_checkpoint_file.h"

#ifdef HAVE_CONFIG_H
#include "zram_config.h"
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

/* Constants, Types ----------------------------------------------------- */

#define POLL_LOSS 10
#define POLL_MAX 200
/* POLL_LOSS=10 and POLL_MAX=200 are good values for the n-queens program */


/* Extern variables ----------------------------------------------------- */

char z_lb_CheckpointPath[PATH_MAX] = "check"; /* name of checkpoint file */
boolean z_lb_traceCheckpoint=1;	/* trace checkpoints */
boolean z_lb_writeCheckpoint=1;	/* write checkpoint files */
boolean z_lb_writeStatistics=1;	/* write statistics to stdout */
boolean z_lb_backupCheckpoint=1;
boolean z_lb_redundantComputation; /* read-only for application */
/* This variable is set to true during the startup phase of the load-
   balancing algorithm when several processors evaluate WorkAndSplit
   for the same argument. */
boolean z_lb_needSplit;	/* read-only for application */
boolean z_lb_needPoll;	/* read-only for application */


/* Static variables ----------------------------------------------------- */

static int z_lb_inited;
#ifndef SEQUENTIAL
static z_NodeP global;
static z_lb_ProcSetP psp;
#endif /* ! SEQUENTIAL */
static FILE *cp_file;
static s_seqP checkpointedNodes;
static s_seqP workList;
static StatArrayT stat;
static int cp_sequence;

/* Procedure prototypes ------------------------------------------------- */

void z_lb_RingSender(void);
void z_lb_parallel(void);

#define WorkAndSplit(global, wu, wl) \
        (stat[stat_wsnonred + z_lb_redundantComputation]++, \
	 psp->lb_WorkAndSplit(global, wu, wl))

#define Split(global, wu, wl) \
        (stat[stat_s]++, psp->lb_Split(global, wu, wl))

#define Work(global, wu) \
        (stat[stat_w]++, psp->lb_Work(global, wu))


/* Procedures ----------------------------------------------------------- */

static void PrintStatistics(void)
{
  char line[100];
  
  if (z_lb_writeStatistics) {
    fflush(stdout);
    sprintf(line, "[%2d]S%12.1f%11lu%11lu%11lu %5lu %5lu %5lu\n",
	    z_myproc, z_Gtime(),
	    stat[stat_w], stat[stat_s], stat[stat_wsnonred],
	    stat[stat_wsred], stat[stat_or], stat[stat_rf]);
    /*printf("[%2d]S WorkAndSplit %ld\n", z_myproc, stat[stat_ws]);*/
    /*printf("[%2d]S Requests generated %ld\n", z_myproc, stat[stat_or]);*/
    /*printf("[%2d]S Requests forwarded %ld\n", z_myproc, stat[stat_rf]);*/
    fprintf(stderr, "%s", line);
    fflush(stderr);
  }
} /* PrintStatistics */

static void Write(void *node)
{
  z_WriteNode(cp_file, (z_NodeP) node);
} /* Write */

static void WriteH(void *nodeh)
{
  z_WriteNode(cp_file, *(z_NodeP*)nodeh);
} /* WriteH */

static void ReadH(void *nodeh)
{
  *(z_NodeP*)nodeh = z_ReadNode(cp_file);
} /* ReadH */

static void ReadCheckpointfile(void)
{
  int num, i, version, checkpoint_end, numprocs;
  
  cp_file = fopen(z_lb_CheckpointPath, "r");
  if (cp_file) {
    z_fread(&version, sizeof(version), 1, cp_file);
    if (version < 2 || version > 4) {
      fprintf(stderr, "*%2d*bad checkpoint file version\n", z_myproc);
      z_assert(z_myproc, "-", __FILE__, __LINE__);
    }
    if (version >= 4) {
      z_fread(&numprocs, sizeof(numprocs), 1, cp_file);
      if (numprocs != z_numprocs) {
	fprintf(stderr, "[%2d]checkpoint file is for %d processors\n",
		z_myproc, numprocs);
	z_assert(z_myproc, "-", __FILE__, __LINE__);
      }
      z_fread(&cp_sequence, sizeof(cp_sequence), 1, cp_file);
    } /* version >= 4 */
    z_fread(&stat, sizeof(stat), 1, cp_file);
    z_fread(&num, sizeof(num), 1, cp_file);
    /*fprintf(stderr, "[%2d]num = %d\n", z_myproc, num);*/
    for (i=0; i<num; i++)
      s_seq_InsertLast(workList, z_ReadNode(cp_file));
    s_seq_ForAll(checkpointedNodes, ReadH);
    if (version >= 3) {
      z_fread(&checkpoint_end, sizeof(checkpoint_end), 1, cp_file);
      if (checkpoint_end != CHECKPOINT_END) {
	fprintf(stderr, "*%2d*bad checkpoint file end\n", z_myproc);
	z_assert(z_myproc, "-", __FILE__, __LINE__);
      }
    }
    fclose(cp_file);
  } else {
    perror(z_lb_CheckpointPath);
    z_assert(z_myproc, "-", __FILE__, __LINE__);
  }
  /*fprintf(stderr, "[%2d]read ok.\n", z_myproc);*/
} /* ReadCheckpointfile */


static void WriteCheckpointfile(void)
{
  int num, numprocs = z_numprocs;
  int version = CHECKPOINT_VERSION, checkpoint_end = CHECKPOINT_END;
  
  if (z_lb_writeCheckpoint) {
    if (z_lb_backupCheckpoint)
      z_check_BackupFile(z_lb_CheckpointPath);
    cp_file = fopen(z_lb_CheckpointPath, "w");
    num = s_seq_numElements(workList);
    cp_sequence ++;
    z_check_WriteHeader(cp_file, version, cp_sequence, numprocs, &stat, num);
    s_seq_ForAll(workList, Write);
    s_seq_ForAll(checkpointedNodes, WriteH);
    fwrite(&checkpoint_end, sizeof(checkpoint_end), 1, cp_file);
    fclose(cp_file);
  }
  PrintStatistics();
} /* WriteCheckpointfile */


/* Sequential work ------------------------------------------------------ */

static void CheckpointSeq(void)
     /* called in the sequential version when z_checkpoint is set */
{
  s_seq_refP ref;

  WriteCheckpointfile();
  fflush(stdout);
  if (z_terminate)
    while (s_seq_numElements(workList)) {
      ref = s_seq_First(workList);
      z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
      s_seq_Delete(workList, ref);
    }
  z_checkpoint = 0;
} /* CheckpointSeq */

static void z_lb_WorkSequential(z_ProcSetNum procsetnum,
				z_lb_GlobalP global)
{
  s_seq_refP ref;
  z_lb_ProcSetP psp;
  z_lb_WorkUnitP workUnit;
  
  assert(z_lb_inited);
  psp = (z_lb_ProcSetP)z_GetProcSet(procsetnum);
  assert(psp->lb_Work || psp->lb_WorkAndSplit);
  z_lb_needPoll = 0;
  while (s_seq_numElements(workList)) {
    ref = s_seq_Last(workList);
    workUnit = (z_lb_WorkUnitP) s_seq_GetObject(workList, ref);
    s_seq_Delete(workList, ref);
    if (psp->lb_Work) {
      Work(global, &workUnit);
      if (workUnit) s_seq_InsertLast(workList, workUnit);
    } else
      WorkAndSplit(global, workUnit, workList);
    if (z_checkpoint) CheckpointSeq();
  }
  if (psp->lb_Fini) psp->lb_Fini();
} /* z_lb_WorkSequential */

static void z_lb_RestartWorkSequential(z_ProcSetNum procsetnum,
				       z_lb_GlobalP global)
{
  /* Sequential enumeration */
  assert(z_lb_inited);
  ReadCheckpointfile();
  if (s_seq_numElements(workList))
    z_lb_WorkSequential(procsetnum, global);
} /* z_lb_RestartWorkSequential */


/* Parallel work -------------------------------------------------------- */

#ifndef SEQUENTIAL

#define LEFT(p)	(((p) + 1) % z_numprocs)

static z_JobNum z_lb_parallelJob;
static z_tagT ringReqTag, ringAnsTag, checkpointTag, backTag;
static boolean wantSendRequest;	/* I want to send a request */
static boolean requestPending;	/* I have sent a request */
static boolean senderQueued;
static boolean broadcastTerm;	/* I want to broadcast a checkpoint message */
static boolean globalCheckpoint;
static boolean terminateAfterCheckpoint; /* terminate after this checkpoint */
static boolean gotBack;		/* I got back an unanswered request */
static int left;		/* send request to the left */
static int *request/*[numprocs]*/;	/* cyclic buffer */
static int *ra/*[numprocs]*/;
static int req_in, req_out;

static void ComputeStartSlice(z_lb_WorkUnitP workUnit)
{
  /* Depends on z_myproc and z_numprocs. */
  s_seq_refP ref;
  int i, size;
  unsigned int my, total, t;

  assert(workUnit);
  if (z_numprocs == 1) {
    s_seq_InsertFirst(workList, workUnit);
    return;
  }
  total = z_numprocs;
  my = z_myproc;
  while (1) {
    assert(my < total);
    if (psp->lb_Split)
      Split(global, workUnit, workList);
    else {
      z_lb_redundantComputation = my > 0;
      WorkAndSplit(global, workUnit, workList);
    }
    /* some work may be done more than once here */
    /* z_Poll(); is not a good idea: load balancing makes computing the
       start slices impossible. */
    size = s_seq_numElements(workList);
    if (size == 0) return;	/* there is no work left for us */
    if (size == 1) {		/* there is work for one of us */
      if (my > 0) {
	ref = s_seq_First(workList);
	z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
	s_seq_Delete(workList, ref);
      }
      return;
    }
    /* size >= 2 */
    if (size > total) {		/* distribute 000 111 22 33 */
      t = size / total;
      i = size % total;
      if (i > my) i = my;
      i += my*t;
      for (; i>0; --i) {
	ref = s_seq_First(workList);
	z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
	s_seq_Delete(workList, ref);
      }
      if (size % total > my) t++; /* elements I get */
      while (s_seq_numElements(workList) > t) {
	ref = s_seq_Last(workList);
	z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
	s_seq_Delete(workList, ref);
      }
      return;
    } /* size > total */
    
    /* size <= total */
    i = my % size;
    my = my / size;
    if (total % size <= i)
      total = total / size;
    else
      total = total / size + 1;
    for (; i>0; --i) {		/* delete elements before mine */
      ref = s_seq_First(workList);
      z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
      s_seq_Delete(workList, ref);
    }
    while (s_seq_numElements(workList) > 1) { /* delete elements after mine */
      ref = s_seq_Last(workList);
      z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
      s_seq_Delete(workList, ref);
    }
    if (total <= 1) return;
    ref = s_seq_First(workList);
    workUnit = (z_lb_WorkUnitP) s_seq_GetObject(workList, ref);
    s_seq_Delete(workList, ref);
  } /* while */
} /* ComputeStartSlice */

static boolean ProcessRequest(void)
{
  /* Called by z_lb_RingSender */
  z_lb_WorkUnitP workUnit;
  s_seq_refP ref;

  if (s_seq_numElements(workList) == 1) { /* split */
    ref = s_seq_First(workList);
    workUnit = (z_lb_WorkUnitP) s_seq_GetObject(workList, ref);
    s_seq_Delete(workList, ref);
    if (psp->lb_Split)
      Split(global, workUnit, workList);
    else
      WorkAndSplit(global, workUnit, workList);
  }
  if (s_seq_numElements(workList) >= 2) { /* send first element */
    ref = s_seq_First(workList);
    workUnit = (z_lb_WorkUnitP) s_seq_GetObject(workList, ref);
    s_seq_Delete(workList, ref);
    /*printf("[%2d]Answering request(%d)\n", z_myproc, request[req_out]);*/
    z_AddTagBuffer(ringAnsTag);
    z_AddNodeBuffer(workUnit);
    z_Isend(request[req_out], "answer");
    z_FreeNode(workUnit);
    return 1;
  }
  return 0;
} /* ProcessRequest */

void z_lb_RingSender(void)
{
  int r, next;
  boolean answered;
  
  assert(senderQueued);
  if (broadcastTerm) {		/* PE 0 broadcasts a checkpoint */
    z_AddTagBuffer(checkpointTag);
    z_AddMINTBuffer(terminateAfterCheckpoint);
    z_Ibroadcast("checkpoint");
    broadcastTerm = 0;
  } else if (req_in != req_out) { /* process a request */
    if (!globalCheckpoint) {
      answered = ProcessRequest();
      if (! answered) {
	z_AddTagBuffer(ringReqTag);
	z_AddMINTBuffer(request[req_out]); /* originator */
	r = ra[req_out];
	z_AddMINTBuffer(r);	/* random processor */
	if (r == z_myproc) next = LEFT(request[req_out]);
	else if (r == left) next = LEFT(left);
	else next = left;
	z_Isend(next, "req");
	stat[stat_rf] ++;
      }
    } else {			/* send back the request */
      z_AddTagBuffer(backTag);
      z_Isend(request[req_out], "req back");
    }
    req_out = (req_out + 1) % z_numprocs;
  } else if (wantSendRequest) {	/* send my own request to a random processor */
    /*    fprintf(stderr,"[%2d]Sending own request\n", z_myproc); */
    if (z_numprocs <= 4)
      r = -1;
    else
      r = (z_myproc + 2 + rand() % (z_numprocs - 2)) % z_numprocs;
    z_AddTagBuffer(ringReqTag);
    z_AddMINTBuffer(z_myproc);
    z_AddMINTBuffer(r);
    if (r >= 0) z_Isend(r, "req");
    else z_Isend(left, "req");
    wantSendRequest = 0;
    requestPending = 1;
    stat[stat_or] ++;
  }
  if (req_in != req_out || wantSendRequest)
    z_QueueSender(z_lb_RingSender);
  else
    senderQueued = 0;
} /* z_lb_RingSender */

static void ActivateSender(void)
{
  if (z_numprocs > 1 && !senderQueued) {
    z_QueueSender(z_lb_RingSender);
    senderQueued = 1;
  }
} /* ActivateSender */

static void RingReqHandler(void)
     /* Called when a request message arrives */
{
  int from;

  from = z_GetMINTBuffer();
  if (from == z_myproc) {	/* everybody has seen my request, forget it */
    assert(requestPending);
    requestPending = 0;
  } else {			/* save the request in the cyclic buffer */
    request[req_in] = from;
    ra[req_in] = z_GetMINTBuffer();
    req_in = (req_in + 1) % z_numprocs;
    ActivateSender();		/* and process it when sending is possible */
  }
} /* RingReqHandler */

static void RingAnsHandler(void)
     /* Called when an answer message arrives, i.e. another processor
      has got my request and sends back a work unit. */
{
  z_lb_WorkUnitP workUnit;

  assert(requestPending);	/* I have sent a request. */
  requestPending = 0;
  assert(! s_seq_numElements(workList)); /* I have no work. */
  workUnit = z_GetNodeBuffer();
  assert(workUnit);
  s_seq_InsertFirst(workList, workUnit);
} /* RingAnsHandler */

static void z_lb_CheckpointHandler(void)
{
  /* Called when a checkpoint message arrives: processor 0 has caught a
     signal and broadcasts a checkpoint message */
  globalCheckpoint = 1;
  terminateAfterCheckpoint = z_GetMINTBuffer();
} /* z_lb_CheckpointHandler */

static void z_lb_BackHandler(void)
     /* Called when a back message arrives: Another processor has sent
	back my request in order to write a checkpoint. */
{
  assert(requestPending);
  requestPending = 0;
  gotBack = 1;
} /* z_lb_BackHandler */

static void GlobalCheckpoint(void)
     /* called on all processors */
{
  s_seq_refP ref;

  if (z_lb_traceCheckpoint)
    printf("[%2d]Checkpoint\n", z_myproc);
  wantSendRequest = 0;		/* do not send a request */
  while (senderQueued || requestPending) z_IdleWithHandler(1);
  /* All processors are synchronized. */
  if (z_lb_traceCheckpoint)
    printf("[%2d]Synchronized\n", z_myproc);
  WriteCheckpointfile();
  if (terminateAfterCheckpoint)
    while (s_seq_numElements(workList)) {
      ref = s_seq_First(workList);
      z_FreeNode((z_NodeP) s_seq_GetObject(workList, ref));
      s_seq_Delete(workList, ref);
    }
  fflush(stdout);
  globalCheckpoint = 0;
  wantSendRequest = !terminateAfterCheckpoint && gotBack;
} /* GlobalCheckpoint */

static void Checkpoint0(void)
{
  /* Called on processor 0 when an ALRM or TERM signal is received */
  z_signal_Block();
  z_printf("Entering Checkpoint0\n");
  terminateAfterCheckpoint = z_terminate;
  broadcastTerm = 1;
  globalCheckpoint = 1;
  ActivateSender();
  GlobalCheckpoint();
  z_checkpoint = 0;
  z_check_saveConfig(cp_sequence);
  z_signal_UnBlock();
} /* Checkpoint0 */

#if defined(cenju3) || (Z_SYNCH_CHECKPOINT)
#define CHECKFREQ 50 
static int checkCount=0;
#endif

static void ParallelMainLoop(z_lb_ProcSetP psp, z_lb_GlobalP global)
{
  boolean terminate, done;
  z_lb_WorkUnitP workUnit;
  s_seq_refP ref;
  int poll_timer, poll_next;

  if (! s_seq_numElements(workList)) { /* we start without work */
    wantSendRequest = 1;
    ActivateSender();
  }
  poll_timer = poll_next = 0;
  do { /* while not terminate */
    if (s_seq_numElements(workList)) { /* I have work to do */
      do {
	assert(!requestPending && !wantSendRequest); /*CRASH*/
	ref = s_seq_Last(workList); /* work on the last list element */
	workUnit = (z_lb_WorkUnitP) s_seq_GetObject(workList, ref);
	s_seq_Delete(workList, ref);
	if (psp->lb_Work) {
	  Work(global, &workUnit);
	  if (workUnit) s_seq_InsertLast(workList, workUnit);
	} else
	  WorkAndSplit(global, workUnit, workList);

	/* Polling can be very expensive (relative to Work or WorkAndSplit). */
	/* So we try to do it not too often. */
	poll_timer ++; /* time since last successful poll */
	if (poll_timer >= poll_next) { /* poll this time */
	  done = z_Poll();
	  if (done) {
	    /*fprintf(stderr, "timer=%d\n", poll_timer);*/
	    poll_timer = 0;
	  } else {
	  }
	  if (poll_timer < POLL_MAX * POLL_LOSS)
	    poll_next = poll_timer + poll_timer / POLL_LOSS;
	  else
	    poll_next = poll_timer + POLL_MAX;
	} /* poll this time */
	
	if (z_myproc == 0) {
#if defined(cenju3) || defined(Z_SYNCH_CHECKPOINT)
	    checkCount ++;
	    if (checkCount > CHECKFREQ) {
		checkCount = 0;
		z_checkpoint = 1;
	    }
#endif
	  if (z_checkpoint) Checkpoint0();
	} else
	  if (globalCheckpoint) GlobalCheckpoint();
      } while (s_seq_numElements(workList));
      /*fprintf(stderr, "[%2d]list empty\n", z_myproc);*/
      if (! terminateAfterCheckpoint) {
	wantSendRequest = 1;
	ActivateSender();
      }
    } /* if */
    terminate = z_IdleWithHandler(requestPending);
    if (! terminate)
      if (z_myproc == 0) {
	if (z_checkpoint) Checkpoint0();
      } else
	if (globalCheckpoint) GlobalCheckpoint();
  } while (! terminate);
} /* ParallelMainLoop */

void z_lb_parallel(void)
{
  /* Parallel enumeration: part running on processing elements. */
  z_lb_WorkUnitP workUnit;
  int psn, i;
  boolean restart;

  assert(z_lb_inited);
  assert(s_seq_numElements(workList) == 0);
  req_in = req_out = 0;
  wantSendRequest = requestPending = senderQueued = 0;
  z_lb_needPoll = 1;
  for (i=stat_last-1; i>=0; --i) stat[i] = 0; /* clear statistics */
  {
    /* Read the message broadcast by the front end in z_lb_WorkParallel.
       It contains global data */
    psn = z_GetMINTBuffer();
    psp = (z_lb_ProcSetP) z_GetProcSet(psn);
    assert(psp->lb_Work || psp->lb_WorkAndSplit);
    assert(psp->lb_Split || psp->lb_WorkAndSplit);
    restart = z_GetMINTBuffer();
    global = z_GetNodeBuffer();
    if (restart) {
      ReadCheckpointfile();
      z_StartMessagePassing();
    } else {
      workUnit = z_GetNodeBuffer();
      assert(workUnit);
      z_StartMessagePassing();
      z_lb_redundantComputation = 0;
      ComputeStartSlice(workUnit);
      z_lb_redundantComputation = 0;
    }
  }

  ParallelMainLoop(psp, global);

  if (psp->lb_Fini) psp->lb_Fini();
  PrintStatistics();
} /* z_lb_parallel */


#ifndef PEonly			/* front end only */

static void z_lb_WorkParallel(z_ProcSetNum procsetnum,
		       z_lb_GlobalP global,
		       z_lb_WorkUnitP workUnit)
{
  /* Parallel enumeration: part running on the front end. */

  if (workUnit) {
    assert(z_frontend);		/* can be called on the front end only */
    assert(z_lb_inited);

    /* Broadcast: procsetnum, *global, *workUnit */
    {
      MINT psn;
      
      psn = procsetnum;
      z_AddMINTBuffer(psn);
      z_AddMINTBuffer(0);	/* do not restart */
      z_AddNodeBuffer(global);
      z_AddNodeBuffer(workUnit);
      z_StartParallelJob(z_lb_parallelJob, NULL);
    }
    
    z_WaitWork();
  }
} /* z_lb_WorkParallel */

static void z_lb_RestartWorkParallel(z_ProcSetNum procsetnum,
				     z_lb_GlobalP global)
{
  /* Parallel enumeration: part running on the front end. */
  assert(z_frontend);		/* can be called on the front end only */
  assert(z_lb_inited);
  z_AddMINTBuffer(procsetnum);
  z_AddMINTBuffer(1);		/* restart */
  z_AddNodeBuffer(global);
  z_StartParallelJob(z_lb_parallelJob, NULL);
  z_WaitWork();
} /* z_lb_RestartWorkParallel */


#endif /* ! PEonly */
#endif /* ! SEQUENTIAL */

/* Procedures ----------------------------------------------------------- */

void z_lb_Work(z_ProcSetNum procsetnum,
	       z_lb_GlobalP global,
	       z_lb_WorkUnitP workUnit,
	       boolean sequential)
{
#ifdef SEQUENTIAL
  s_seq_InsertFirst(workList, workUnit);
  z_lb_WorkSequential(procsetnum, global);
#else /* ! SEQUENTIAL */
  if (sequential) {
    s_seq_InsertFirst(workList, workUnit);
    z_lb_WorkSequential(procsetnum, global);
  } else
    z_lb_WorkParallel(procsetnum, global, workUnit);
#endif /* ! SEQUENTIAL */
} /* z_lb_Work */


void z_lb_RestartWork(z_ProcSetNum procsetnum,
		      z_lb_GlobalP global,
		      boolean sequential)
{
#ifdef SEQUENTIAL
  z_lb_RestartWorkSequential(procsetnum, global);
#else /* ! SEQUENTIAL */
  if (sequential)
    z_lb_RestartWorkSequential(procsetnum, global);
  else
    z_lb_RestartWorkParallel(procsetnum, global);
#endif /* ! SEQUENTIAL */
} /* z_lb_RestartWork */



/* Initialisation ------------------------------------------------------- */

void z_lb_InstallCheckpointed(z_NodeP *nodeh)
     /* Add nodeh to the list of nodes which will be written to the
	checkpoint file */
{
  assert(z_lb_inited);
  s_seq_InsertLast(checkpointedNodes, nodeh);
} /* z_lb_InstallCheckpointed */

void z_lb_Init(void)
/* Initialize this module */
{
  /* called on every processor by initproc() */
  if (! z_lb_inited) {
    z_lb_inited = 1;
    z_io_Init();
#ifndef SEQUENTIAL
    z_mp_Init();
    z_lb_parallelJob = z_InstallJob(z_lb_parallel);
    ringReqTag = z_InstallMessage(RingReqHandler);
    ringAnsTag = z_InstallMessage(RingAnsHandler);
    checkpointTag = z_InstallMessage(z_lb_CheckpointHandler);
    backTag = z_InstallMessage(z_lb_BackHandler);
    left = LEFT(z_myproc);
    request = (int *) malloc(z_numprocs * sizeof(*request));
    ra = (int *) malloc(z_numprocs * sizeof(*ra));
#endif /* ! SEQUENTIAL */
    workList = s_seq_New(s_seq_default);
    checkpointedNodes = s_seq_New(s_seq_default);

  }
} /* z_lb_Init */

/* Issues:

   - compare sequence number of all checkpoint files

   */

